import 'dart:async';
import 'dart:math';

import 'package:clean_android_design/clean_android_design.dart';
import 'package:clean_android_design/src/dialog/save_or_abort_dialog.dart';
import 'package:clean_android_design/src/services/update_scaffold_widget_service_interface.dart';
import 'package:clean_android_design/src/services/models_interfaces.dart';
import 'package:clean_android_design/src/structure/simple_scaffold_menu_button_widget.dart';
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter/material.dart';

import 'default_scaffold_widget.dart';
import 'icon_scaffold_menu_button_widget.dart';
import 'scaffold_page_widget.dart';

class DefaultUpdateScaffoldWidget<T extends UpdateableDatabaseItem<T>>
    extends StatefulWidget {
  final IUpdateScaffoldWidgetService<T> databaseCollection;
  final T Function() createNewItem;
  final String updateableItemId;
  final void Function(BuildContext context, UpdateItemFrame<T> updateItemFrame)
      beforeSave;
  final Future<void> Function(
      BuildContext context, UpdateItemFrame<T> updateItemFrame) beforeSaveAsync;
  final void Function(BuildContext context, DocumentReference itemReference)
      afterInsert;
  final void Function(BuildContext context, UpdateItemFrame<T> updateItemFrame)
      onAbort;
  final void Function(BuildContext context, T item) prepareNewItem;
  final void Function(Validator<T> validator) initValidator;
  final int timeoutMilliseconds;

  final List<Widget> Function(
          BuildContext context, UpdateItemFrame<T> updateItemFrame)
      menuButtonsBuilder;
  final List<Widget> Function(
          BuildContext context, UpdateItemFrame<T> updateItemFrame)
      editMenuButtonsBuilder;
  final List<Widget> menuButtons;
  final List<Widget> editMenuButtons;

  final List<Widget> Function(
          BuildContext context, UpdateItemFrame<T> updateItemFrame)
      appBarButtonsBuilder;
  final List<Widget> Function(
          BuildContext context, UpdateItemFrame<T> updateItemFrame)
      editAppBarButtonsBuilder;
  final List<Widget> appBarButtons;
  final List<Widget> editAppBarButtons;
  final Widget waitingTitle;

  final bool hasEditButton;
  final bool isDefaultEditable;

  final Widget Function(
      BuildContext context, UpdateItemFrame<T> updateItemFrame) titleBuilder;
  final List<ScaffoldPageWidget<T>> Function(
      BuildContext context,
      UpdateItemFrame<T> updateItemFrame,
      ReadOnlyValidator<T> scaffoldValidator) updatePagesBuilder;
  final Widget Function(
      BuildContext context,
      UpdateItemFrame<T> updateItemFrame,
      ReadOnlyValidator<T> validator) builder;
  final IHelpWidget helpWidget;

  const DefaultUpdateScaffoldWidget({
    Key key,
    @required this.createNewItem,
    this.updateableItemId,
    this.menuButtons,
    this.hasEditButton = true,
    this.updatePagesBuilder,
    this.editMenuButtons,
    this.titleBuilder,
    this.beforeSave,
    this.prepareNewItem,
    this.onAbort,
    this.beforeSaveAsync,
    this.afterInsert,
    this.isDefaultEditable = false,
    this.helpWidget,
    this.builder,
    this.initValidator,
    this.menuButtonsBuilder,
    this.editMenuButtonsBuilder,
    this.appBarButtonsBuilder,
    this.editAppBarButtonsBuilder,
    this.appBarButtons,
    this.editAppBarButtons,
    @required this.databaseCollection,
    this.waitingTitle,
    this.timeoutMilliseconds = 4000,
  })  : assert(titleBuilder != null),
        assert(beforeSave == null || beforeSaveAsync == null),
        assert(databaseCollection != null),
        assert(createNewItem != null),
        super(key: key);

  @override
  _DefaultUpdateScaffoldWidgetState<T> createState() =>
      _DefaultUpdateScaffoldWidgetState<T>();
}

class _DefaultUpdateScaffoldWidgetState<T extends UpdateableDatabaseItem<T>>
    extends State<DefaultUpdateScaffoldWidget<T>> {
  UpdateItemFrame<T> updateItemFrame;
  bool isNewItem;

  StreamSubscription<T> streamSubscriptionDocument;
  StreamSubscription<bool> streamSubscriptionInternet;

  Key validatePageKey;
  Key mainKey;
  Validator<T> validator;

  List<ScaffoldPageWidget<T>> pages;

  bool waitingForUpdate = false;
  bool gotUpdateButStillEditing = false;
  int itemOrigin;

  static const int itemOriginWaiting = 0;
  static const int itemOriginServer = 1;
  static const int itemOriginCache = 2;
  static const int itemOriginNone = 3;
  static const int itemOriginNew = 4;

  @override
  void initState() {
    super.initState();

    updateItemFrame = UpdateItemFrame();
    updateItemFrame.setNewItem(widget.createNewItem());
    isNewItem = widget.updateableItemId == null;
    updateItemFrame.setEditable(isNewItem || widget.isDefaultEditable);

    if (isNewItem || widget.isDefaultEditable)
      Future.delayed(
          Duration.zero, () => LockedNavigatorWidget.setLocked(true));

    if (isNewItem) {
      itemOrigin = itemOriginNew;
      try {
        widget.prepareNewItem?.call(context, updateItemFrame.editableItem);
      } catch (e) {}
    }

    validatePageKey = null;
    mainKey = Key(Random().nextInt(10000).toString());

    if (!isNewItem) {
      itemOrigin = itemOriginWaiting;

      try {
        streamSubscriptionDocument = widget.databaseCollection
            .documentStream(widget.updateableItemId)
            ?.listen(snapshotListener);
      } catch (e) {}
      var getItem = () async {
        try {
          var item = await widget.databaseCollection
              .document(widget.updateableItemId, source: Source.server);

          if (item == null) {
            item = await widget.databaseCollection
                .document(widget.updateableItemId, source: Source.cache);

            if (item == null) return itemOriginNone;

            snapshotListener(item);
            return itemOriginCache;
          } else {
            snapshotListener(item);

            return itemOriginServer;
          }
        } catch (e) {
          return itemOriginNone;
        }
      };

      getItem()
          .timeout(Duration(milliseconds: widget.timeoutMilliseconds),
              onTimeout: () => itemOriginNone)
          .then((origin) {
        if (mounted)
          setState(() {
            itemOrigin = origin;
          });
      });
    }

    streamSubscriptionInternet = CleanAndroidDesignSettings.connectionChange
        ?.listen((_) => setState(() {}));
  }

  @override
  Widget build(BuildContext context) {
    if (validator == null && widget.initValidator != null) {
      validator = Validator<T>();
      widget.initValidator(validator);
    }
    pages = widget.updatePagesBuilder
        ?.call(context, updateItemFrame, validator?.readOnly());

    if (itemOrigin == itemOriginWaiting)
      return LoadingScaffoldWidget(
        title: widget.waitingTitle ?? Text("Details"),
        showBackArrow: true,
      );

    if (itemOrigin == itemOriginNone)
      return DefaultScaffoldWidget(
        title: Text("Offline"),
        showBackArrow: true,
        child: Column(
          mainAxisSize: MainAxisSize.min,
          mainAxisAlignment: MainAxisAlignment.start,
          children: [
            Container(height: 80),
            Icon(
              Icons.wifi_off,
              size: 50,
              color: Colors.grey[600],
            ),
            DescriptionText(
                "Daten konnten nicht geladen werden, da zurzeit keine Internetvebindung besteht"),
          ],
        ),
      );

    return DefaultScaffoldWidget(
      key: mainKey,
      title: widget.titleBuilder(context, updateItemFrame),
      showBackArrow: true,
      onWillPop: _onWillPop,
      menuButtons: _getMenuButtons(),
      appBarButtons: _getAppBarButtons(),
      updatePages: pages,
      child:
          widget.builder?.call(context, updateItemFrame, validator?.readOnly()),
      defaultPageKey: validatePageKey,
      helpWidget: widget.helpWidget,
      showLoading: waitingForUpdate,
    );
  }

  List<Widget> _getAppBarButtons() {
    return ((updateItemFrame.isEditable
            ? (widget.editAppBarButtons ??
                widget.editAppBarButtonsBuilder?.call(context, updateItemFrame))
            : null) ??
        widget.appBarButtons ??
        widget.appBarButtonsBuilder?.call(context, updateItemFrame) ??
        <Widget>[])
      ..add(itemOrigin == itemOriginCache
          ? IconButton(
              icon: Icon(
                Icons.signal_wifi_off,
              ),
              tooltip: "Offline Daten",
              onPressed: null,
            )
          : (itemOrigin == itemOriginWaiting
              ? IconButton(
                  icon: Container(
                    width: 18,
                    height: 18,
                    child: CircularProgressIndicator(
                      valueColor: AlwaysStoppedAnimation<Color>(
                          CleanAndroidDesignDefaultColors.defaultColor),
                      strokeWidth: 2,
                    ),
                  ),
                  onPressed: null,
                  tooltip: "Lade Daten von Server",
                )
              : Container()));
  }

  List<Widget> _getMenuButtons() {
    if (!widget.hasEditButton)
      return widget.menuButtons ??
          widget.editMenuButtonsBuilder?.call(context, updateItemFrame) ??
          <Widget>[];

    if (updateItemFrame.isEditable)
      return <Widget>[
        SimpleScaffoldMenuButtonWidget(
          text: Text("Abbrechen"),
          onTap: () => _abortEditing(),
          tooltip: "Bearbeitung abbrechen",
        ),
        SimpleScaffoldMenuButtonWidget(
          text: Text("Speichern"),
          onTap: () => _validate(),
          tooltip: "Änderungen speichern",
        )
      ]..addAll(
          widget.editMenuButtons ??
              widget.editMenuButtonsBuilder?.call(context, updateItemFrame) ??
              widget.menuButtons ??
              widget.menuButtonsBuilder?.call(context, updateItemFrame) ??
              <Widget>[],
        );
    else
      return <Widget>[
        IconScaffoldMenuButtonWidget(
          icon: Icons.edit,
          text: Text("Bearbeiten"),
          disabled: !CleanAndroidDesignSettings.isConnected(),
          onTap: _tryStartEditing,
          tooltip: "Eigenschaften bearbeiten",
        )
      ]..addAll(widget.menuButtons ??
          widget.menuButtonsBuilder?.call(context, updateItemFrame) ??
          <Widget>[]);
  }

  void _tryStartEditing() async {
    FocusScope.of(context).unfocus();
    LoadingWidget.show(context);

    var check = () async {
      try {
        if (await CleanAndroidDesignSettings.checkIsConnected()) {
          if (CleanAndroidDesignSettings.isLocked != null &&
              await CleanAndroidDesignSettings.isLocked(
                updateItemFrame.editableItem.getTypeName(),
                updateItemFrame.editableItem.id,
              )) {
            return -2;
          } else {
            return 0;
          }
        } else {
          return -1;
        }
      } catch (e) {
        return -1;
      }
    };

    var result = await check().timeout(
        Duration(milliseconds: widget.timeoutMilliseconds),
        onTimeout: () => -3);

    Navigator.pop(context);

    if (result == 0) {
      // successfull
      CleanAndroidDesignSettings.setLock?.call(
        updateItemFrame.editableItem.getTypeName(),
        updateItemFrame.editableItem.id,
      );
      setState(() {
        updateItemFrame.setEditable(true);
      });
      LockedNavigatorWidget.setLocked(true);
    } else if (result == -1 || result == -3) {
      SnackBarController.showText("Keine Internetverbindung!");
    } else if (result == -2) {
      await showCustomDialog(
          context,
          DefaultDialog(
            title: Text("Dieser Eintrag wird zurzeit bearbeitet"),
            buttons: [
              DefaultDialogButton(
                titel: Text("Verstanden"),
              )
            ],
          ));
    }
  }

  Future<bool> _onWillPop() async {
    FocusScope.of(context).unfocus();

    if (updateItemFrame.isEditable) {
      if (updateItemFrame.editableItem.equals(updateItemFrame.originalItem)) {
        LockedNavigatorWidget.setLocked(false);
        return true;
      } else {
        bool result = false;
        await showCustomDialog(
            context,
            SaveOrAbortDialog(
              onAbort: () {
                result = true;
                CleanAndroidDesignSettings.removeLock?.call(
                  updateItemFrame.editableItem.getTypeName(),
                  updateItemFrame.editableItem.id,
                );
                LockedNavigatorWidget.setLocked(false);
              },
              onSave: () {
                result = false;
                _validate();
              },
            ));
        return result;
      }
    } else
      return true;
  }

  void snapshotListener(T item) {
    if (!mounted || item == null) return;
    updateItemFrame.setNewItem(item);
    if (waitingForUpdate) {
      if (updateItemFrame.isEditable) {
        gotUpdateButStillEditing = true;
      } else {
        waitingForUpdate = false;
      }
    }
    setState(() {});
  }

  @override
  void dispose() {
    super.dispose();

    streamSubscriptionInternet?.cancel();
    streamSubscriptionDocument?.cancel();
  }

  void _abortEditing() {
    FocusScope.of(context).unfocus();

    try {
      widget.onAbort?.call(context, updateItemFrame);

      CleanAndroidDesignSettings.removeLock?.call(
        updateItemFrame.editableItem.getTypeName(),
        updateItemFrame.editableItem.id,
      );
    } catch (e) {}

    setState(() {
      if (isNewItem) {
        Navigator.pop(context);
      } else {
        updateItemFrame.setEditable(false);
      }

      LockedNavigatorWidget.setLocked(false);
    });
  }

  bool _validate() {
    FocusScope.of(context).unfocus();

    var result = validator?.isValid(updateItemFrame.editableItem);
    if (validator == null || result.isValid) {
      validatePageKey = null;

      _saveItem();
      return true;
    } else {
      if (pages != null) {
        setState(() {
          validatePageKey = result.firstInvalidPage;
          mainKey = Key(Random().nextInt(10000).toString());
        });
        Future.delayed(Duration(milliseconds: 100), () {
          var result = validator.isValid(updateItemFrame.editableItem);
          if (result.firstInvalidNode != null) {
            FocusScope.of(context).unfocus();
            FocusScope.of(context).requestFocus(result.firstInvalidNode);
          }
        });
      } else {
        setState(() {});
        Future.delayed(Duration(milliseconds: 100), () {
          FocusScope.of(context).unfocus();
          FocusScope.of(context).requestFocus(result.firstInvalidNode);
        });
      }
      return false;
    }
  }

  void _saveItem() async {
    try {
      setState(() => waitingForUpdate = true);

      if (!(await CleanAndroidDesignSettings.checkIsConnected())) {
        await showCustomDialog(
            context,
            DefaultDialog(
              dismissible: true,
              title: Text(
                  "Es besteht zurzeit keine Internetverbindung. Bitte versuche es später erneut."),
              buttons: [DefaultDialogButton(titel: Text("Okay"))],
            ));

        setState(() => waitingForUpdate = false);
        return;
      }

      if (widget.beforeSave != null)
        widget.beforeSave(context, updateItemFrame);

      if (widget.beforeSaveAsync != null)
        await widget.beforeSaveAsync(context, updateItemFrame);

      gotUpdateButStillEditing = false;
      if (isNewItem) {
        var itemRef =
            await widget.databaseCollection.add(updateItemFrame.editableItem);
        widget.afterInsert?.call(context, itemRef);
        Navigator.pop(context);
      } else {
        var subItemsChanged =
            (updateItemFrame.editableItem is HasDatabaseSubItems &&
                    updateItemFrame.originalItem is HasDatabaseSubItems)
                ? !(updateItemFrame.editableItem as HasDatabaseSubItems)
                    .equalSubItems(
                        updateItemFrame.originalItem as HasDatabaseSubItems)
                : false;
        await widget.databaseCollection.update(updateItemFrame.editableItem,
            supressSubItemUpdate: !subItemsChanged);

        setState(() {
          if (gotUpdateButStillEditing) {
            gotUpdateButStillEditing = false;
            waitingForUpdate = false;
          }
          updateItemFrame.setEditable(false);
        });
      }

      CleanAndroidDesignSettings.removeLock?.call(
        updateItemFrame.editableItem.getTypeName(),
        updateItemFrame.editableItem.id,
      );

      LockedNavigatorWidget.setLocked(false);
    } catch (e) {
      setState(() {
        gotUpdateButStillEditing = false;
        waitingForUpdate = false;
      });
      await showCustomDialog(
          context,
          DefaultDialog(
            dismissible: true,
            title: Text(
                "Es ist ein Fehler aufgetreten. Möglicherweise besteht zurzeit keine Internetverbindung."),
            buttons: [DefaultDialogButton(titel: Text("Okay"))],
          ));
    }
  }
}

class UpdateItemFrame<T extends UpdateableDatabaseItem<T>> {
  bool _isEditable = false;

  // if isEditable = false, all three are the same
  T _currentDBItem;
  T _originalItem;
  T _editableItem;

  void setNewItem(T item) {
    _currentDBItem = item;
    if (!isEditable) {
      _originalItem = item;
      _editableItem = item;
    }
  }

  void setEditable(bool isEditable) {
    _isEditable = isEditable;
    if (isEditable) {
      _editableItem = _currentDBItem.copy();
      _originalItem = _currentDBItem.copy();
    } else {
      _editableItem = _currentDBItem;
      _originalItem = _currentDBItem;
    }
  }

  bool get isEditable => _isEditable;
  T get currentDBItem => _currentDBItem;
  T get originalItem => _originalItem;
  T get editableItem => _editableItem;
}
